unit ERGNDIFPClientUnit;

interface

uses
  SysUtils, Classes, ClientSockets,ScktComp, WinSock;

type
  TNDIFPClient = class; // forward declaration

  T3dMarker = packed record
    x,y,z, reliability : Single;
  end;
  P3DMarker = ^T3DMarker;

  T3DData = packed record
    FrameNumber : LongWord;
    TimeStamp : UInt64;
    NumMarkers : longint;
    Position : packed array of T3DMarker;
  end;

  TOdauData = packed record
    FrameNumber : LongWord;
    TimeStamp : UInt64;
    NumChannels : longint;
    Voltage : packed array of Single;
  end;

  T6DBody = packed record
    Q0, Qx, Qy, Qz, X, Y, Z, Error : Single;
  end;

  T6DData = packed record
    FrameNumber : LongWord;
    TimeStamp : UInt64;
    NumMarkers : longint;
    CurrentFrame : packed array of T6DBody;
  end;

  TDataOptions = (do3dData, doOdauData, doForce, do6DData, doEvent);

  TDataOptionsSet = set of TDataOptions;

  TConnectHandler = procedure (Sender : TNDIFPClient) of object;
  TDataHandler = procedure (Sender : TNDIFPClient; DataOptions : TDataOptionsSet) of object;
  TErrorHandler = procedure (Sender : TObject; ErrorString : AnsiString) of object;
  TNetworkConnectionErrorHandler = procedure (Sender: TObject; Socket: TCustomWinSocket;
      ErrorEvent: TErrorEvent; var ErrorCode: Integer) of object;

  TNDIFPClient = class(TComponent)
  private
    FOnData: TDataHandler;
    FPort: Integer;
    FHost: String;
    FOnNDIFPError: TErrorHandler;
    FOnNetworkConnectionError: TNetworkConnectionErrorHandler;
    FOnConnect: TConnectHandler;
    //FTimer: TPerformanceTimer;
    //FMemo: TMemo;
    FOnDisconnect: TConnectHandler;
    FOnConnecting: TConnectHandler;
    procedure SetOnData(const Value: TDataHandler);
    procedure SetHost(const Value: String);
    procedure SetPort(const Value: Integer);
    procedure SetOnNDIFPError(const Value: TErrorHandler);
    procedure SetOnNetworkConnectionError(
      const Value: TNetworkConnectionErrorHandler);
    procedure SetOnConnect(const Value: TConnectHandler);
    function GetRigidBodyData: T6DData;
    //procedure SetTimer(const Value: TPerformanceTimer);
    //procedure SetMemo(const Value: TMemo);
    procedure SetOnConnecting(const Value: TConnectHandler);
    procedure SetOnDisconnect(const Value: TConnectHandler);
    { Private declarations }
  protected
    { Protected declarations }
    theData : AnsiString;
    thePacket : AnsiString;
    PacketLength : LongWord;
    FOdauData : TOdauData;
    FMarkerData : T3Ddata;
    FRigidBodyData : T6DData;
    ClientSocket : TNagleClientSocket;
    procedure Loaded; override;
    procedure SendCommand(S : AnsiString);
    procedure OnSocketConnecting(Sender: TObject;Socket: TCustomWinSocket);
    procedure OnSocketConnect(Sender: TObject;Socket: TCustomWinSocket);
    procedure OnSocketDisconnect(Sender: TObject;Socket: TCustomWinSocket);
    procedure OnSocketError(Sender: TObject; Socket: TCustomWinSocket;
      ErrorEvent: TErrorEvent; var ErrorCode: Integer);
    procedure OnSocketRead(Sender: TObject; Socket: TCustomWinSocket);
    function GetOdauData : TOdauData;
    function GetMarkerData : T3dData;
    property OnNetworkConnectionError : TNetworkConnectionErrorHandler read FOnNetworkConnectionError write SetOnNetworkConnectionError;
  public
    { Public declarations }
    t : Extended;
    Property OdauData : TOdauData read GetOdauData;
    property MarkerData : T3Ddata read GetMarkerData;
    property RigidBodyData : T6DData read GetRigidBodyData;
    constructor Create(AOwner : TComponent); override;
    destructor Destroy; override;
    procedure About;
    procedure Connect;
    procedure Disconnect;
    procedure StartStream;
    procedure StopStream;
    procedure StartCollection;
    procedure StopCollection;
  published
    { Published declarations }
    //property Timer : TPerformanceTimer read FTimer write SetTimer;
    //property Memo : TMemo read FMemo write SetMemo;
    property OnConnecting: TConnectHandler read FOnConnecting write SetOnConnecting;
    property OnConnect : TConnectHandler read FOnConnect write SetOnConnect;
    property OnDisconnect: TConnectHandler read FOnDisconnect write SetOnDisconnect;
    property OnNDIFPError : TErrorHandler read FOnNDIFPError write SetOnNDIFPError;
    property OnData : TDataHandler read FOnData write SetOnData;
    property Host : String read FHost write SetHost;
    property Port : Integer read FPort write SetPort;
  end;

procedure Register;

implementation

uses Dialogs, ABOUTERGNDIFPClientUnit;

procedure Register;
begin
  RegisterComponents('Optotrak', [TNDIFPClient]);
end;

{ TNDIFPClient }

procedure TNDIFPClient.About;
begin
  // show Aboutbox containing Version Info, Author and some more things
  AboutBox.ShowModal;
end;

procedure TNDIFPClient.Connect;
begin
  ClientSocket.Active := true;
end;

constructor TNDIFPClient.Create(AOwner: TComponent);
begin
  inherited Create(AOWner);
  ClientSocket := nil;
  port := 3020;
end;

destructor TNDIFPClient.Destroy;
begin
  ClientSocket.Active := false;
  ClientSocket.Free;
  ClientSocket := nil;
  inherited;
end;

procedure TNDIFPClient.Disconnect;
begin
  if ClientSocket.Active then
  begin
    SendCommand('StreamFrames Stop');
    SendCommand('Bye');
  end;
  ClientSocket.Active := false;
end;

function TNDIFPClient.GetMarkerData: T3dData;
begin
  result := FMarkerData;
end;

function TNDIFPClient.GetOdauData: TOdauData;
begin
  result := FOdauData;
end;

function TNDIFPClient.GetRigidBodyData: T6DData;
begin
  result := FRigidBodyData;
end;

procedure TNDIFPClient.Loaded;
begin
  inherited;
  ClientSocket := TNagleClientSocket.Create(Self);
  ClientSocket.Host := host;
  ClientSocket.port := port;
  ClientSocket.Usenagle := false;
  ClientSocket.OnError := OnSocketError;
  ClientSocket.OnRead := OnSocketRead;
  ClientSocket.OnConnecting := OnSocketConnecting;
  ClientSocket.OnConnect := OnSocketConnect;
  ClientSocket.OnDisconnect := OnSocketDisconnect;
end;

procedure TNDIFPClient.OnSocketConnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  SendCommand('Version 1.0');
  SendCommand('SetByteOrder LittleEndian');
  SendCommand('SendParameters AllFrames All');
  if assigned(OnConnect) then
  begin
    OnConnect(Self);
  end;
end;

procedure TNDIFPClient.OnSocketConnecting(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  if assigned(OnConnecting) then
  begin
    OnConnecting(Self);
  end;
end;

procedure TNDIFPClient.OnSocketDisconnect(Sender: TObject;
  Socket: TCustomWinSocket);
begin
  if assigned(OnDisconnect) then
  begin
    OnDisconnect(Self);
  end;
end;

procedure TNDIFPClient.OnSocketError(Sender: TObject; Socket: TCustomWinSocket;
  ErrorEvent: TErrorEvent; var ErrorCode: Integer);
begin
  if assigned(OnNetworkConnectionError) then
  begin
    OnNetworkConnectionError(Self, Socket, ErrorEvent, ErrorCode);
  end else
  begin
    MessageDlg('SocketError: ' + inttostr(ErrorCode), mtError, [mbOK], 0);
  end;
end;

procedure TNDIFPClient.OnSocketRead(Sender: TObject; Socket: TCustomWinSocket);
var DataOptions : TDataOptionsSet;

  procedure processDataFrame(s : AnsiString);
  var count : longint;
      i, j : integer;
      size : LongWord;
      dataType : byte;
      framenr : LongWord;
      TimeStamp : UInt64;
      cnt : longword;

      procedure Process3Ddata(s1 : AnsiString);
      begin
        FMarkerData.FrameNumber := FrameNr;
        FMarkerData.TimeStamp := TimeStamp;
        FMarkerData.NumMarkers := cnt;
        SetLength(FMarkerData.Position, cnt);
        move(s1[1], FMarkerData.Position[0], Length(s1));
      end;

      procedure ProcessAnalogdata(s1 : AnsiString);
      begin
        FODauData.FrameNumber := FrameNr;
        FOdauData.TimeStamp := TimeStamp;
        FODauData.NumChannels := cnt;
        SetLength(FOdauData.voltage, cnt);
        Move(s1[1], FOdauData.voltage[0], length(s1));
      end;

      procedure Process6Ddata(s1 : AnsiString);
      begin
        FRigidBodyData.FrameNumber := FrameNr;
        FRigidBodyData.TimeStamp := TimeStamp;
        FRigidBodyData.NumMarkers := cnt;
        SetLength(FRigidBodyData.CurrentFrame, cnt);
        move(s1[1], FRigidBodyData.CurrentFrame[0], Length(s1));
      end;


  begin
    count := 0;
    for i := 4 downto 1 do // big endian
    begin
      count := count * 256 + ord(s[i]);
    end;
    s := copy(s,5, length(s) - 4);
    for j := 1 to count do
    begin
      size := 0;
      for i := 4 downto 1 do
      begin
        size := size * 256 + ord(s[i]);
      end;
      datatype := Ord(s[5]);
      framenr := 0;
      for i := 12 downto 9 do
      //for i := 9 to 12 do
      begin
        framenr := framenr * 256 + ord(s[i]);
      end;
      timestamp := 0;
      for i := 20 downto 13 do
      //for i := 13 to 20 do
      begin
        timestamp := timestamp * 256 + ord(s[i]);
      end;
      cnt := 0;
      for i := 24 downto 21 do
      begin
        cnt := cnt * 256 + ord(s[i]);
      end;
      case datatype of
        1:
          begin
            // 3d data
            DataOptions := DataOptions + [do3DData];
            Process3Ddata(Copy(s,25,size - 24));
          end;
        2:
          begin
            // analog odau data
            DataOptions := DataOptions + [doOdauData];
            ProcessAnalogdata(Copy(s,25,size - 24));
          end;
        3:
          begin
            // force
            DataOptions := DataOptions + [doForce];
          end;
        4:
          begin
            // 6D data
            DataOptions := DataOptions + [do6DData];
            Process6Ddata(Copy(s,25,size - 24));
          end;
        5:
          begin
            // Event data
            DataOptions := DataOptions + [doEvent];
          end;
        else
          begin
            raise Exception.Create('Unsupported Dataframe of type ' + inttostr(dataType));
          end;
      end; // end of case statement
      s := copy(s, size + 1, length(s) - size);
    end;
  end;

  procedure ProcessPacket(s : AnsiString);
  begin
    case Ord(s[8]) of
    0:
      begin
        // Error package
        s := copy(s,9,length(s) - 8);
        if Assigned(OnNDIFPError) then
        begin
          OnNDIFPError(Self, s);
        end else
        begin
          raise Exception.Create(s);
        end;
      end;
    1:
      begin
        // command string, command succeeded
      end;
    2:
      begin
        // XML
      end;
    3:
      begin
        // Data frame
        //memo2.lines.add('Data frame received..');
        DataOptions := [];
        ProcessDataFrame(copy(s,9,length(s) - 8));
        if assigned(OnData) then  // Dit moet zo in het te maken component komen.
        begin
          OnData(Self, DataOptions);
        end;//}
      end;
    4:
      begin
        // No data.
        //This packet type has no body. It indicates that a
        //measurement has finished or is not yet started.
      end;
    5:
      begin
        // Complete C3D file
        // A complete C3D file sent from the server.
      end;
    else
      begin
        raise Exception.Create('Unrecognized data type: ' + inttostr(Ord(s[8])));
      end;
    end;
  end;

  procedure ProcessData;
  var i : integer;
  begin
    if PacketLength = 0 then
    begin
      if Length(theData) >= 4 then
      begin
        for i := 1 to 4 do
        begin
          PacketLength := PacketLength * 256 + Ord(theData[i]);
        end;
      end;
    end;
    if (PacketLength > 0) and (Length(theData) >= PacketLength) then
    begin
      DataOptions := [];
      thePacket := Copy(theData, 1, PacketLength);
      // Reset the received buffer
      theData := Copy(theData, PacketLength + 1, Length(theData) - PacketLength);
      PacketLength := 0;
      //process the packet
      ProcessPacket(thePacket);
      ProcessData;  // go recursive on any remaining information in 'theData'
    end;
  end;

begin
  //t := Timer.Seconds;
  theData := theData + Socket.ReceiveText;
  ProcessData;
  //t := Timer.Seconds - t;
  //memo.lines.add(inttostr(trunc(t * 1000000 + 0.5)));
end;

procedure TNDIFPClient.SendCommand(S: AnsiString);
var s2 : AnsiString;
    l : longint;
    i : integer;
begin
  l := length(s) + 4 + 4 + 1;
  s2 := '';
  for i := 1 to 4 do
  begin
    s2 := Chr(l mod 256) + s2;
    l := l div 256;
  end;
  s2 := s2 + #0 + #0 + #0 + #1;
  s2 := s2 + s + #0;
  ClientSocket.Socket.SendBuf(s2[1],length(s2));
end;


procedure TNDIFPClient.SetHost(const Value: String);
begin
  FHost := Value;
  if ClientSocket <> nil  then
  begin
    if not ClientSocket.Active then
    begin
      ClientSocket.Host := Host;
    end else
    begin
      raise Exception.Create('Cannot set property TNDIFPClient.Host when connected.');
    end;
  end;
end;

{procedure TNDIFPClient.SetMemo(const Value: TMemo);
begin
  FMemo := Value;
end;// }

procedure TNDIFPClient.SetOnConnect(const Value: TConnectHandler);
begin
  FOnConnect := Value;
end;

procedure TNDIFPClient.SetOnConnecting(const Value: TConnectHandler);
begin
  FOnConnecting := Value;
end;

procedure TNDIFPClient.SetOnData(const Value: TDataHandler);
begin
  FOnData := Value;
end;

procedure TNDIFPClient.SetOnDisconnect(const Value: TConnectHandler);
begin
  FOnDisconnect := Value;
end;

procedure TNDIFPClient.SetOnNDIFPError(const Value: TErrorHandler);
begin
  FOnNDIFPError := Value;
end;

procedure TNDIFPClient.SetOnNetworkConnectionError(
  const Value: TNetworkConnectionErrorHandler);
begin
  FOnNetworkConnectionError := Value;
end;

procedure TNDIFPClient.SetPort(const Value: Integer);
begin
  FPort := Value;
  if clientsocket <> nil then
  begin
    if not ClientSocket.Active then
    begin
      ClientSocket.Port := Port;
    end else
    begin
      Raise Exception.Create('Cannot set property TNDIFPClient.Port when connected.');
    end;
  end;
end;

{procedure TNDIFPClient.SetTimer(const Value: TPerformanceTimer);
begin
  FTimer := Value;
end;//}

procedure TNDIFPClient.StartCollection;
begin
  MessageDlg('Unimplemented method: '+#13+#10+'StartCollection' + #13+#10 + 'This method should trigger the OptoToolStrober, but we don''t have one yet.', mtWarning, [mbOk],0);
end;

procedure TNDIFPClient.StartStream;
begin
  //SendCommand('StreamFrames Frequency:50 3D Analog');
  SendCommand('StreamFrames AllFrames All');
end;

procedure TNDIFPClient.StopCollection;
begin
  MessageDlg('Unimplemented method: '+#13+#10+'StopCollection' + #13+#10 + 'This method should trigger the OptoToolStrober, but we don''t have one yet.', mtWarning, [mbOk],0);
end;

procedure TNDIFPClient.StopStream;
begin
  SendCommand('StreamFrames Stop');
end;

end.
